/*
 * $Id: _spi_common.h,v 1.3 2009/11/02 18:50:27 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: _spi_common.h,v $
 * Revision 1.3  2009/11/02 18:50:27  clivewebster
 * Added revision log
 *
 * ===========
 * _spi_common.h
 *
 *  Created on: 23-Jun-2009
 *      Author: Clive Webster
 *
 *  The abstract class for all SPI interfaces whether they are implemented in hardware or software
 */

#ifndef _SPI_COMMON_H_
#define _SPI_COMMON_H_

#include "libdefs.h"
#include "iopin.h"

typedef enum{
	SPI_DATA_ORDER_MSB=0,	// Data is sent MSB first
	SPI_DATA_ORDER_LSB=1	// Data is sent LSB first
} SPI_DATA_ORDER;

typedef enum{
	SPI_MODE_0 = 0,	// CPOL=0, CPHA =0
	SPI_MODE_1 = 1,	// CPOL=0, CPHA =1
	SPI_MODE_2 = 2,	// CPOL=1, CPHA =0
	SPI_MODE_3 = 3	// CPOL=1, CPHA =1
} SPI_MODE;

// Set the clock speed to the CPU speed divided by 2, 4, 8 etc
typedef enum{
	SPI_CLOCK_DIV2 = 2,
	SPI_CLOCK_DIV4 = 4,
	SPI_CLOCK_DIV8 = 8,
	SPI_CLOCK_DIV16 = 16,
	SPI_CLOCK_DIV32 = 32,
	SPI_CLOCK_DIV64 = 64,
	SPI_CLOCK_DIV128 = 128
} SPI_CLOCK;


struct s_spi_common;

// ------- Define the structure for an attached SPI Device --------
typedef struct s_spi_device{
	struct s_spi_common* bus;	// The spi bus it is connected to
	SPI_MODE		mode;		// The mode used by this device - Normally SPI_MODE_0
	SPI_DATA_ORDER	order;		// The bit order used by this device - Normally SPI_DATA_ORDER_MSB
	const IOPin *   select;		// The pin used to select the device, can be null if there is only one device
	boolean			initialised;// Has the device been initialised
} SPI_DEVICE;

#define MAKE_SPI_DEVICE(mode,order,select) { null,mode,order,select,FALSE}

// A datatype that store a list of pointers to SPI_DEVICEs
typedef SPI_DEVICE* PROGMEM  SPI_DEVICE_LIST;




//-------- Define the variables stored for each SPI bus  ---
typedef struct s_spi_commmon {
	const struct PROGMEM c_spi*	class;	// The type of implementation
	SPI_DEVICE_LIST* const devices;		// The list of attached devices
	size_t			numDevices;			// The number of devices
	SPI_CLOCK		clock;				// The clock value
	SPI_MODE		mode;				// The mode
	SPI_DATA_ORDER	order;				// The bit order
	boolean			initialised;		// Has it been initialised
} _SPI;
#define _MAKE_SPI(class,devices) {class,devices,(uint8_t)(sizeof(devices)/sizeof(SPI_DEVICE*)),SPI_CLOCK_DIV2,SPI_MODE_0,SPI_DATA_ORDER_MSB, FALSE}


// --- Define methonds available to all SPI_DEVICEs ---
static __inline__ _SPI* __spiDeviceGetBus(struct s_spi_common* arg){
	return (_SPI*)arg;
}

// Select or deselect this device
#define spiDeviceSelect(device, active) _spiDeviceSelect(&((device)->_device_),active)
void _spiDeviceSelect(SPI_DEVICE* device, boolean active);

//------- Define methods available to all SPI interfaces ---
#define spiSendWord(spi,data) _spiSendWord(&((spi)->_spi_),data)
uint16_t _spiSendWord(_SPI* spi, uint16_t data);

#define spiGetClock(spi) (spi)->_spi_.clock
#define _spiGetClock(spi) (spi)->clock

#define spiGetMode(spi) (spi)->_spi_.mode
#define _spiGetMode(spi) (spi)->mode

#define spiGetDataOrder(spi) (spi)->_spi_.order
#define _spiGetDataOrder(spi) (spi)->order


//---------------- Define virtual methods -------------------
// Initialise the spi interface as a master or slave
#define spiInit(spi,master) _spiInit(&((spi)->_spi_),master)
void _spiInit(_SPI* spi, boolean master);

// Turn off the SPI interface
#define spiOff(spi) _spiOff(&((spi)._spi_))
void _spiOff(_SPI* spi);

// Set the SPI clock speed
#define spiSetClock(spi,clock) _spiSetClock(&((spi)->_spi_),clock)
void _spiSetClock(_SPI* spi,SPI_CLOCK clock);

// Set the bit order
#define spiSetDataOrder(spi,order) _spiSetDataOrder(&((spi)->_spi_),order)
void _spiSetDataOrder(_SPI* spi,SPI_DATA_ORDER order);

// Set the operating mode
#define spiSetMode(spi,mode) _spiSetMode(&((spi)->_spi_),mode)
void _spiSetMode(_SPI* spi,SPI_MODE mode);

// Send a byte
// For a Master it will send the byte to the slave and return the received byte from the slave
// For a Slave  it will send the byte  back to the master and return the next byte from the master
#define spiSendByte(spi,data) _spiSendByte(&((spi)->_spi_),data)
uint8_t _spiSendByte(_SPI* spi, uint8_t data);

typedef struct PROGMEM c_spi{
	void     (*init)  (_SPI* spi,boolean master);		// Initialise the device
	void     (*off)   (_SPI* spi);						// Turn off the device
	void     (*clock) (_SPI* spi, SPI_CLOCK clock);		// Set clock speed
	void 	 (*order) (_SPI* spi,SPI_DATA_ORDER order);	// Set the bit order
	void 	 (*mode)  (_SPI* spi,SPI_MODE mode);		// Set the SPI mode
	uint8_t	 (*send)  (_SPI* spi,uint8_t data);			// Send a byte
} SPI_CLASS;
// Create macro for defining an SPI implementation
#define MAKE_SPI_CLASS(init, off, clock, order, mode,send) { init, off, clock, order, mode,send }

#endif /* _SPI_COMMON_H_ */
