/*
 * $Id: _uart_common.c,v 1.6 2009/11/02 18:50:27 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: _uart_common.c,v $
 * Revision 1.6  2009/11/02 18:50:27  clivewebster
 * Added revision log
 *
 * ===========
 * _uart_common.c
 *
 *  Created on: 04-May-2009
 *      Author: Clive Webster
 */

#include "_uart_common.h"
#include "core.h"


// Virutal function - start sending the given byte
void _uartStartXmit(UART* uart,uint8_t txData){
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*,uint8_t) = (void (*)(UART*,uint8_t))pgm_read_word(&class->startXmit);
	fn(uart,txData);
}

// We are starting to transmit a block
static void _uartTransmitBlockStart(UART* uart){
	uart->readyTx = FALSE;				// indicate we are sending
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*) = (void (*)(UART*))pgm_read_word(&class->startXmitMode);
	if(fn){
		fn(uart);
	}
}

// We have finished transmitting a block
static void _uartTransmitBlockEnd(UART* uart){
	uart->readyTx = TRUE;				// indicate we are not longer sending
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*) = (void (*)(UART*))pgm_read_word(&class->endXmitMode);
	if(fn){
		fn(uart);
	}
}



// A byte has been received
void __uartReceiveService(UART* uart, uint8_t c){
	uart->rxChar = c;
	// if there's a user function to handle this receive event
	if(uart->rxFunc){
		// call it and pass the received data
		uart->rxFunc(c);
	}else if( uart->rxBuffer ){
		// there is a receive buffer so try to append it
		if(bufferPut(__uartGetRxBuffer(uart),c)==FALSE){
			// track overflows
			uart->rxOverflow=TRUE;
			setError(UART_RECEIVE_OVERFLOW);
		}
	}
}

// We have finished transmitting the current byte
void __uartTransmitService(UART* uart){
	if(uart->sendingBuffer){
		// there is a transmit buffer
		uint8_t c;

		CRITICAL_SECTION_START;
		if(bufferGet(__uartGetTxBuffer(uart),&c)){
			// send the next byte
			_uartStartXmit(uart,c);
		}else{
			// all done
			_uartTransmitBlockEnd(uart);		// we have finished sending
			uart->sendingBuffer = FALSE;		// we are not sending a buffer
		}
		CRITICAL_SECTION_END;
	}else{
		CRITICAL_SECTION_START;
		// indicate transmit complete, back to ready
		_uartTransmitBlockEnd(uart);				// we have finished sending
		CRITICAL_SECTION_END;
	}
}


// Gets a single byte from the uart receive buffer.
// Returns the byte, or -1 if no byte is available (getchar-style).
int __uartGetByte(UART* uart){
	// get single byte from receive buffer (if available)
	uint8_t c;
	int rtn;

	cBuffer* buffer = __uartGetRxBuffer(uart);
	if(buffer!=null){
		// There is a receive buffer
		if(bufferGet(buffer,&c))
			rtn = c;
		else
			rtn = -1;
	}else{
		CRITICAL_SECTION_START;
		rtn = uart->rxChar;		// get the last received char
		uart->rxChar = -1;		// and reset it
		CRITICAL_SECTION_END;
	}
	return rtn;
}


// Sends a single byte over the uart.
uint8_t _uartSendByte(UART* uart, uint8_t txData){
	if(uart->txBuffer){
		// Wait if the buffer is full
		while(bufferIsFull(uart->txBuffer)){
			breathe();
		}

		CRITICAL_SECTION_START;
		if(!uart->sendingBuffer){
			// Not already transmitting so start sending the buffer
			uart->sendingBuffer = TRUE;			// we are sending a buffer
			_uartTransmitBlockStart(uart);			// indicate we are sending
			_uartStartXmit(uart,txData);		// put the byte
		}else{
			// add char to the output queue
			bufferPut(uart->txBuffer,txData);
		}
		CRITICAL_SECTION_END;

	}else{
		// There is no xmit buffer - just send now

		// Wait till sending has stopped
		while(!uart->readyTx){
			breathe();
		}
		CRITICAL_SECTION_START;
		_uartTransmitBlockStart(uart);			// indicate we are sending
		_uartStartXmit(uart,txData);		// put the byte
		CRITICAL_SECTION_END;
	}
	return txData;
}

void _uartSetBaudRate(UART* uart, BAUD_RATE baud){
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*,BAUD_RATE) = (void (*)(UART*,BAUD_RATE))pgm_read_word(&class->setBaudRate);
	fn(uart,baud);
}

void _uartOff(UART* uart){
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*) = (void (*)(UART*))pgm_read_word(&class->off);

	// Wait till everything has been sent
	INTERRUPTABLE_SECTION_START;
	while(__uartIsBusy(uart));
	INTERRUPTABLE_SECTION_END;

	fn(uart);
}

// Is this a single wire interface
boolean _uartIsOneWire(const UART* uart){
	return (uart->rx_pin == uart->tx_pin) ? TRUE : FALSE;
}


void _uartInit(UART* uart, BAUD_RATE baud){
	const UART_CLASS* class = uart->class;
	void (*fn)(UART*,BAUD_RATE) = (void (*)(UART*,BAUD_RATE))pgm_read_word(&class->init);
	cli();

	// For one wire Make the pins into inputs with pullups enabled on transmit
	if(_uartIsOneWire(uart)){
		pin_make_input(uart->tx_pin,TRUE);
		pin_make_input(uart->rx_pin,FALSE);
	}

	// Initialise local variable
	uart->readyTx = TRUE;
	uart->rxOverflow = FALSE;
	uart->sendingBuffer = FALSE;
//	uart->rxFunc = null;

	fn(uart,baud);
	sei();
}
