/*
 * $Id: i2c_master.c,v 1.5 2009/11/02 18:58:48 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: i2c_master.c,v $
 * Revision 1.5  2009/11/02 18:58:48  clivewebster
 * Added revision log
 *
 * ===========
 * i2c_master.c
 *
 *  Created on: 10-Sep-2009
 *      Author: Clive Webster
 */

// TWS4:3 set - ACK has been received back
#define TW_MT_SLA_ACK				0x18

// TWS6 set
#define TW_MR_SLA_ACK				0x40

// defines and constants
#define TWCR_CMD_MASK		0x0F
#define TWSR_STATUS_MASK	0xF8



#include "i2c_master.h"
#include "libdefs.h"

// Include the device definitions
#include <avr/io.h>

static inline void i2cSendStart(void){
	// send start condition
	outb(TWCR, (inb(TWCR)&TWCR_CMD_MASK)|BV(TWINT)|BV(TWSTA));
}

static inline void i2cSendStop(void){
	// transmit stop condition
	// leave with TWEA on for slave receiving
	outb(TWCR, (inb(TWCR)&TWCR_CMD_MASK)|BV(TWINT)|BV(TWEA)|BV(TWSTO));
}

static inline void i2cWaitForComplete(void){
	// wait for i2c interface to complete operation
	while( !(inb(TWCR) & BV(TWINT)) );
}

static inline void i2cSendByte(uint8_t data){
	// save data to the TWDR
	outb(TWDR, data);
	// begin send
	outb(TWCR, (inb(TWCR)&TWCR_CMD_MASK)|BV(TWINT));
}

static inline void i2cReceiveByte(boolean ackFlag){
	// begin receive over i2c
	if( ackFlag ){
		// ackFlag = TRUE: ACK the recevied data
		outb(TWCR, (inb(TWCR)&TWCR_CMD_MASK)|BV(TWINT)|BV(TWEA));
	}else{
		// ackFlag = FALSE: NACK the recevied data
		outb(TWCR, (inb(TWCR)&TWCR_CMD_MASK)|BV(TWINT));
	}
}

static inline uint8_t i2cGetReceivedByte(void){
	// retieve received data byte from i2c TWDR
	return( inb(TWDR) );
}

extern const IOPin* I2C_SCL;
extern const IOPin* I2C_SDA;

// Initialiase I2C hardware
void __attribute__ ((constructor)) _private_i2cMasterInit(void){

#if defined (__AVR_ATmega640__)
#	define SCL_PORT PORTD
#	define SCL_DDR  DDRD
#	define SCL_PIN  PD0
#	define SDA_PORT PORTD
#	define SDA_DDR  DDRD
#	define SDA_PIN  PD1
#elif defined (__AVR_ATmega8__)
#	define SCL_PORT PORTC
#	define SCL_DDR  DDRC
#	define SCL_PIN  PC5
#	define SDA_PORT PORTC
#	define SDA_DDR  DDRC
#	define SDA_PIN  PC4
#elif defined (__AVR_ATmega168__)
#	define SCL_PORT PORTC
#	define SCL_DDR  DDRC
#	define SCL_PIN  PC5
#	define SDA_PORT PORTC
#	define SDA_DDR  DDRC
#	define SDA_PIN  PC4
#elif defined (__AVR_ATmega32__)
#	define SCL_PORT PORTC
#	define SCL_DDR  DDRC
#	define SCL_PIN  PC0
#	define SDA_PORT PORTC
#	define SDA_DDR  DDRC
#	define SDA_PIN  PC1
#elif defined (__AVR_ATmega2560__)
#	define SCL_PORT PORTD
#	define SCL_DDR  DDRD
#	define SCL_PIN  PD0
#	define SDA_PORT PORTD
#	define SDA_DDR  DDRD
#	define SDA_PIN  PD1
#elif defined (__AVR_ATmega328P__)
#	define SCL_PORT PORTC
#	define SCL_DDR  DDRC
#	define SCL_PIN  PORTC5
#	define SDA_PORT PORTC
#	define SDA_DDR  DDRC
#	define SDA_PIN  PORTC4
#else
#	error I2C pins not defined
#endif
	// Make the 2 wires into inputs
	cbi(SCL_DDR,SCL_PIN);
	cbi(SDA_DDR,SDA_PIN);
	// and turn on the pullups
	sbi(SCL_PORT,SCL_PIN);
	sbi(SDA_PORT,SDA_PIN);

	// set i2c bit rate to 100KHz
	i2cSetBitrate(100);

	// enable TWI (two-wire interface)
	sbi(TWCR, TWEN);

	// set state
	i2cState = I2C_IDLE;

	// enable TWI interrupt and slave address ACK
//	sbi(TWCR, TWIE);
//	sbi(TWCR, TWEA);

	// enable interrupts
//	sei();
}


// Send out bytes immediately
boolean i2cMasterSend(uint8_t deviceAddr, size_t length, const uint8_t* data){
	uint8_t retval = TRUE;

	// disable TWI interrupt
	cbi(TWCR, TWIE);

	// send start condition
	i2cSendStart();
	i2cWaitForComplete();

	// send device address with write
	i2cSendByte( deviceAddr & 0xFE );
	i2cWaitForComplete();

	// check if device is present and live and got an ACK back
	if( inb(TWSR) == TW_MT_SLA_ACK)	{
		// send data
		while(length)
		{
			i2cSendByte( *data++ );
			i2cWaitForComplete();
			length--;
		}
	}else{
		// device did not ACK it's address,
		// data will not be transferred
		// return error
		retval = FALSE;
	}

	// transmit stop condition
	// leave with TWEA on for slave receiving
	i2cSendStop();
	while( !(inb(TWCR) & BV(TWSTO)) );

	// enable TWI interrupt
	sbi(TWCR, TWIE);

	return retval;
}

boolean i2cMasterReceive(uint8_t deviceAddr, size_t length, uint8_t *data){
	boolean retval = TRUE;

	// disable TWI interrupt
	cbi(TWCR, TWIE);

	// send start condition
	i2cSendStart();
	i2cWaitForComplete();

	// send device address with read
	i2cSendByte( deviceAddr | 0x01 );
	i2cWaitForComplete();

	// check if device is present and live
	if( inb(TWSR) == TW_MR_SLA_ACK)
	{
		// accept receive data and ack it
		while(length > 1)
		{
			i2cReceiveByte(TRUE);
			i2cWaitForComplete();
			*data++ = i2cGetReceivedByte();
			// decrement length
			length--;
		}

		// accept receive data and nack it (last-byte signal)
		i2cReceiveByte(FALSE);
		i2cWaitForComplete();
		*data++ = i2cGetReceivedByte();
	}
	else
	{
		// device did not ACK it's address,
		// data will not be transferred
		// return error
		retval = FALSE;
	}

	// transmit stop condition
	// leave with TWEA on for slave receiving
	i2cSendStop();

	// enable TWI interrupt
	sbi(TWCR, TWIE);

	return retval;
}
