/*
 * $Id: libdefs.h,v 1.9 2010/01/24 14:35:32 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: libdefs.h,v $
 * Revision 1.9  2010/01/24 14:35:32  clivewebster
 * *** empty log message ***
 *
 * Revision 1.8  2009/11/02 19:00:09  clivewebster
 * Added revision log
 *
 * ===========
 * libdefs.h
 *
 *  Created on: 11-Mar-2009
 *      Author: Clive Webster
 */

#ifndef LIBDEFS_H_
#define LIBDEFS_H_

// include common WinAVR stuff
#include <inttypes.h>
#include <stdint.h>
#include <avr/common.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

// create a type for boolean
typedef int8_t boolean;

// create a type to store timer us ticks
// maximum is 0xFFFFFFFF or 4294967295us = 4294 seconds or 71 minutes
typedef uint32_t TICK_COUNT;

// Create a type to store a loop count
typedef uint32_t LOOP_COUNT;

// Define some hardware data types
typedef uint16_t PORT; // A port number eg _SFR_MEM_ADDR(PORTA)
typedef uint8_t PIN;  // A pin numnber eg 0...7


#ifndef null
  // Define a value for 'null'
# define null ((void*)0)
#endif


#ifndef CRITICAL_SECTION_START
#define CRITICAL_SECTION_START	unsigned char _sreg = SREG; cli()
#define CRITICAL_SECTION_END    SREG = _sreg
#endif


#ifndef INTERRUPTABLE_SECTION_START
#define INTERRUPTABLE_SECTION_START	unsigned char _sreg = SREG; sei()
#define INTERRUPTABLE_SECTION_END    SREG = _sreg
#endif


// Read a byte from ROM or RAM
#define READMEMBYTE(rom,char_ptr)	((rom)?(pgm_read_byte(char_ptr)):(*char_ptr))

// --------------- avrlibdefs.h ------------------------
// Code compatibility to new AVR-libc
// outb(), inb(), inw(), outw(), BV(), sbi(), cbi(), sei(), cli()
#ifndef outb
	#define	outb(addr, data)	addr = (data)
#endif
#ifndef inb
	#define	inb(addr)			(addr)
#endif
#ifndef outw
	#define	outw(addr, data)	addr = (data)
#endif
#ifndef inw
	#define	inw(addr)			(addr)
#endif

#ifndef BV
	#define BV(bit)			(1<<(bit))
#endif
#ifndef cbi
	#define cbi(reg,bit)	reg &= ~(BV(bit))
#endif
#ifndef sbi
	#define sbi(reg,bit)	reg |= (BV(bit))
#endif
#ifndef cli
	#define cli()			__asm__ __volatile__ ("cli" ::)
#endif
#ifndef sei
	#define sei()			__asm__ __volatile__ ("sei" ::)
#endif
#ifndef nop
	#define nop()			__asm__ __volatile__ ("nop" ::)
#endif



// use this for packed structures
// (this is seldom necessary on an 8-bit architecture like AVR,
//  but can assist in code portability to AVR)
#define GNUC_PACKED __attribute__((packed))


// MIN/MAX/ABS macros
#define MIN(a,b)			((a<b)?(a):(b))
#define MAX(a,b)			((a>b)?(a):(b))
#define ABS(x)				((x>0)?(x):(-x))

// Clamp a value to be in a given range
#define CLAMP(val, min, max) MAX(MIN(val,max),min)

// constants
#define PI		3.14159265359


// ----------------- avrlibtypes.h -------------------------------
// Include standard files from AVRLIB
#ifndef WIN32
	// true/false defines
	#define FALSE	0
	#define TRUE	-1
	#define false	FALSE
	#define true	TRUE
#endif

// datatype definitions macros
//typedef uint8_t  u08;
//typedef  int8_t  s08;
//typedef uint16_t u16;
//typedef  int16_t s16;
//typedef uint32_t u32;
//typedef  int32_t s32;
//typedef unsigned long long u64;
//typedef   signed long long s64;

// maximum value that can be held
// by unsigned data types (8,16,32bits)
#define MAX_U08	255U
#define MAX_U16	65535UL
#define MAX_U32	4294967295UL

// maximum values that can be held
// by signed data types (8,16,32bits)
#define MIN_S08	-128
#define MAX_S08	127
#define MIN_S16	-32768
#define MAX_S16	32767
#define MIN_S32	-2147483648
#define MAX_S32	2147483647

//#ifndef WIN32
	// more type redefinitions
//	typedef uint8_t   BOOL;
//	typedef uint8_t	  BYTE;
//	typedef uint16_t  WORD;
//	typedef uint32_t  DWORD;

//	typedef BYTE  UCHAR;
//	typedef WORD  UINT;
//	typedef WORD  USHORT;
//	typedef DWORD ULONG;

//	typedef int8_t	CHAR;
//	typedef int16_t	INT;
//	typedef int32_t	LONG;
//#endif

typedef int8_t ERROR_CODE;

// A type to store an ADC channel number ouutput by a2d:ADC_NUMBER_TO_CHANNEL
typedef uint8_t ADC_CHANNEL;

typedef uint32_t BAUD_RATE;


#define UNUSED_PORT _SFR_MEM8(0X00)

// Define a baud rate that represents 'the highest possible' - I know it looks like a weird value
// but it is trapped and the baud rate registe pair is set to '1'
#define BAUD_RATE_MAX ((BAUD_RATE)MAX_U32)

#if defined(BUILDING_LIBRARY)
extern const uint32_t cpu_speed;
extern const uint16_t cpu_speed_div_1000;
extern const uint16_t cpu_speed_div_1000000;
#else
const uint32_t cpu_speed = F_CPU;
const uint16_t cpu_speed_div_1000 = F_CPU / 1000U;
const uint16_t cpu_speed_div_1000000 = F_CPU / 1000000U;
#endif


typedef struct s_portmask{
	const PORT		port;
	const PIN 		mask;
} PORT_MASK;

/******************************************************

	 Define a structure to hold the definition
	 of an I/O Port

*******************************************************/

typedef struct s_ioport{
	const PORT		port;
	const PORT		ddr;
	const PORT		pin;
}IOPort;
/******************************************************

	 Define a structure to hold the definition
	 of an I/O Pin ie a Port and a Pin number

*******************************************************/
typedef struct s_IOPin {
	const IOPort* port;	// The port
	PIN pin;			// The ie BV(PB2)
} IOPin;



/* ---------------------------------------------------------------------------

Define the name for each of the possible timer modes 0...16

---------------------------------------------------------------------------- */
typedef enum {
	TIMER_MODE_NORMAL=0,
	TIMER_MODE_PWM8_PHASE_CORRECT,
	TIMER_MODE_PWM9_PHASE_CORRECT,
	TIMER_MODE_PWM10_PHASE_CORRECT,
	TIMER_MODE_CTC_OCR,
	TIMER_MODE_PWM8_FAST,
	TIMER_MODE_PWM9_FAST,
	TIMER_MODE_PWM10_FAST,
	TIMER_MODE_PWM_PHASE_FREQ_ICR,
	TIMER_MODE_PWM_PHASE_FREQ_OCR,
	TIMER_MODE_PWM_PHASE_CORRECT_ICR,
	TIMER_MODE_PWM_PHASE_CORRECT_OCR,
	TIMER_MODE_CTC_ICR,
	TIMER_MODE_13_RESVD,
	TIMER_MODE_PWM_FAST_ICR,
	TIMER_MODE_PWM_FAST_OCR
} TIMER_MODE;
#define MAX_MODES 16

/*
	Declare a datatype that can store a list of modes, one bit per TIMER_MODE.
*/
typedef uint16_t TIMER_MODES;

// The set of PWM modes
#define PWM_MODES ((TIMER_MODES)( \
	BV(TIMER_MODE_PWM8_PHASE_CORRECT) | \
	BV(TIMER_MODE_PWM9_PHASE_CORRECT) | \
	BV(TIMER_MODE_PWM10_PHASE_CORRECT) | \
	BV(TIMER_MODE_PWM8_FAST) | \
	BV(TIMER_MODE_PWM9_FAST) | \
	BV(TIMER_MODE_PWM10_FAST) | \
	BV(TIMER_MODE_PWM_PHASE_FREQ_ICR) | \
	BV(TIMER_MODE_PWM_PHASE_FREQ_OCR) | \
	BV(TIMER_MODE_PWM_PHASE_CORRECT_ICR) | \
	BV(TIMER_MODE_PWM_PHASE_CORRECT_OCR) | \
	BV(TIMER_MODE_PWM_FAST_ICR) | \
	BV(TIMER_MODE_PWM_FAST_OCR) \
))

// The set of ICR modes
#define ICR_MODES ((TIMER_MODES)( \
	BV(TIMER_MODE_PWM_PHASE_FREQ_ICR) | \
	BV(TIMER_MODE_PWM_PHASE_CORRECT_ICR) | \
	BV(TIMER_MODE_CTC_ICR) | \
	BV(TIMER_MODE_PWM_FAST_ICR)  \
))

// The set of OCR modes
#define OCR_MODES ((TIMER_MODES)( \
	BV(TIMER_MODE_CTC_OCR) | \
	BV(TIMER_MODE_PWM_PHASE_FREQ_OCR) | \
	BV(TIMER_MODE_PWM_PHASE_CORRECT_OCR) | \
	BV(TIMER_MODE_PWM_FAST_OCR)  \
))

#define modeIsPWM(mode) ( ((TIMER_MODES)(BV(mode))) & PWM_MODES)
#define modeIsICR(mode) ( ((TIMER_MODES)(BV(mode))) & ICR_MODES)
#define modeIsOCR(mode) ( ((TIMER_MODES)(BV(mode))) & OCR_MODES)

#define TIMER_NO_MODES 0

// Define bits if all modes are supported
#define TIMER_ALL_MODES ((TIMER_MODES)( \
	BV(TIMER_MODE_NORMAL)|\
	BV(TIMER_MODE_PWM8_PHASE_CORRECT)|\
	BV(TIMER_MODE_PWM9_PHASE_CORRECT)|\
	BV(TIMER_MODE_PWM10_PHASE_CORRECT)|\
	BV(TIMER_MODE_CTC_OCR)|\
	BV(TIMER_MODE_PWM8_FAST)|\
	BV(TIMER_MODE_PWM9_FAST)|\
	BV(TIMER_MODE_PWM10_FAST)|\
	BV(TIMER_MODE_PWM_PHASE_FREQ_ICR)|\
	BV(TIMER_MODE_PWM_PHASE_FREQ_OCR)|\
	BV(TIMER_MODE_PWM_PHASE_CORRECT_ICR)|\
	BV(TIMER_MODE_PWM_PHASE_CORRECT_OCR)|\
	BV(TIMER_MODE_CTC_ICR)|\
	BV(TIMER_MODE_PWM_FAST_ICR)|\
	BV(TIMER_MODE_PWM_FAST_OCR)\
))

// Define bits for less capable timers
#define TIMER_3BIT_MODES (BV(TIMER_MODE_NORMAL)| BV(TIMER_MODE_PWM8_PHASE_CORRECT)|BV(TIMER_MODE_CTC_OCR)| \
		BV(TIMER_MODE_PWM8_FAST)|BV(TIMER_MODE_PWM_PHASE_CORRECT_OCR)|BV(TIMER_MODE_PWM_FAST_OCR))

#define TIMER_2BIT_MODES (BV(TIMER_MODE_NORMAL)| BV(TIMER_MODE_PWM8_PHASE_CORRECT)|BV(TIMER_MODE_CTC_OCR)| \
		BV(TIMER_MODE_PWM8_FAST))


/**
	Define the constant info for a timer compare unit
	This information will be held in PROGMEM
*/
typedef struct s_timer_compare{
	void* const 		pgm_data;					// The address of the volatile data in SRAM of type TimerDataCompare*
	const uint8_t		pgm_timer;					// The timer number 0...NUMBER_OF_TIMERS
	PORT_MASK			pgm_intenable;				// The port and bit mask that is used to turn on/off compare interrupts
	const PORT			pgm_threshold;				// The port to set the compare value, For a sixteen bit port should be the low one ie OCR1AL
	PORT_MASK			pgm_intpending;				// The port and bit mask that is used to indicate an interrupt is pending
	PORT_MASK			pgm_com;					// The port and bit number used to store COMxy0
	const IOPin * PROGMEM	pgm_pwm;				// The pwm pin if any
} TimerCompare;

// Define a callback for TimerCompare events
typedef void (*TimerCompareCallback)(const TimerCompare *timer_compare, void* data);

/**
	Define the volatile info for a timer compare unit
	This information will be held in Flash SRAM
*/
typedef struct s_timer_compare_data{
	TimerCompareCallback compare_callback; 	// The callback function for this timer	when it compare is met
	void* 				compare_data;		// The data passed to the callback routine
    volatile TICK_COUNT overflow; 			// The timer overflow value incremented via interrupts. Thats why its volatile
} TimerDataCompare;


#define MAKE_TIMER_COMPARE_DATA() {0,null,0}
#define MAKE_TIMER_COMPARE(data,timer,port,mask,threshold, intport,intmask, comport,combit,pwm) \
	{&data,timer, {_SFR_MEM_ADDR(port),BV(mask)},_SFR_MEM_ADDR(threshold), \
	{_SFR_MEM_ADDR(intport),BV(intmask)}, \
	{_SFR_MEM_ADDR(comport),combit}, \
	pwm \
	}

/**
	Define the constant info for a timer
	This information will be held in PROGMEM
*/
typedef struct s_timer{
	void* const   	pgm_data;			// The address of the volatile data in SRAM of type TimerData*
	const PORT	  	pgm_counter;		// The port that counts up from the prescaler eg TCNT0
	const PORT	  	pgm_prescaler;		// The port used for the prescaler
	const boolean 	pgm_sixteenBit;		// TRUE if it is a 16 bit timer
	const boolean 	pgm_rtc;			// TRUE if this uses the TimerRTCPrescaleFactor values
	const uint8_t 	pgm_numCompare;		// The number of compare channels
	const TIMER_MODES pgm_modes;		// The modes supported by this timer
	PORT_MASK 		pgm_wgm[4];			// Up to 4 WGM bits
	const TimerCompare* pgm_compares;    // The compare units
	PORT_MASK	  	pgm_overflowint;	// The port/mask for overflow interrupt enable
	PORT_MASK	  	pgm_overflowreq;	// The port/mask used to indicate an overflow interrupt is pending
	const PORT		pgm_icr;			// Input compare register
	PORT_MASK	  	pgm_captureint;		// The port/mask for capture interrupt enable
	PORT_MASK	  	pgm_capturereq;		// The port/mask used to indicate a capture interrupt is pending
	PORT_MASK	  	pgm_captureedge;	// The port/mask used to select rising falling edge for capture input
	const IOPin*	pgm_capturepin;		// The pin used for input capture, or null if not available
} Timer;

// Declare a data type for the timer interrrupt callback function
typedef void (*TimerCallback)(const Timer *timer, void* data);


/**
	Define the volatile info for a timer.
	This information will be held in Flash SRAM
*/
typedef struct s_timer_data{
	uint16_t			prescale_value;		// The value used in the prescaler. use timerSetPrecaler to change
    volatile TICK_COUNT overflow; 			// The timer overflow value incremented via interrupts. Thats why its volatile
	TimerCallback	  	overflow_callback; 	// The callback function for this timer	when its counter overflow
	void* 				overflow_data;		// The data passed to the callback overflow routine
	TimerCallback		capture_callback;	// The callback function for this timer	when its capture pin is raised
	void* 				capture_data;		// The data passed to the callback capture routine
} TimerData;

#define MAKE_TIMER_DATA(prescale) {prescale,0,null,null,null,null}
#define MAKE_TIMER(data, counter,prescaler,sixteenBit,rtc,modes, \
	wgm0port, wgm0mask, wgm1port, wgm1mask, wgm2port, wgm2mask, wgm3port, wgm3mask, \
    compares,ovrintport,ovrintmask,ovrreqport,ovrreqmask,icr, \
    capintport,capintmask,capreqport,capreqmask, capedgeport,capedgemask,\
    incapin \
	) \
	{&data, _SFR_MEM_ADDR(counter), _SFR_MEM_ADDR(prescaler), sixteenBit, rtc, sizeof(compares)/sizeof(TimerCompare),modes, \
    	{ \
			{_SFR_MEM_ADDR(wgm0port),BV(wgm0mask)}, \
			{_SFR_MEM_ADDR(wgm1port),BV(wgm1mask)}, \
			{_SFR_MEM_ADDR(wgm2port),BV(wgm2mask)}, \
			{_SFR_MEM_ADDR(wgm3port),BV(wgm3mask)}  \
    	},\
    	compares,\
	{_SFR_MEM_ADDR(ovrintport),BV(ovrintmask)},\
	{_SFR_MEM_ADDR(ovrreqport),BV(ovrreqmask)}, \
	_SFR_MEM_ADDR(icr),\
	{_SFR_MEM_ADDR(capintport),BV(capintmask)},\
	{_SFR_MEM_ADDR(capreqport),BV(capreqmask)}, \
	{_SFR_MEM_ADDR(capedgeport),BV(capedgemask)}, \
	incapin \
	}

// Define the signature of code that is the destination of an rprintf output
typedef uint8_t (*Writer)(uint8_t byte);
#define MAKE_WRITER(name) uint8_t name(uint8_t byte)


#endif /* LIBDEFS_H_ */
