/*
 * $Id: motorPWM.c,v 1.7 2009/11/02 19:00:09 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: motorPWM.c,v $
 * Revision 1.7  2009/11/02 19:00:09  clivewebster
 * Added revision log
 *
 * ===========
 * motorPWM.c
 *
 *  Created on: 24-Mar-2009
 *      Author: Clive Webster
 */

#include "motorPWM.h"
#include "errors.h"
#include "iopin.h"
#include "timer.h"
#include "core.h"

// Callback - for when the speed has been set
static void setSpeed(__ACTUATOR *actuator, DRIVE_SPEED speed){
	MOTOR* motor = (MOTOR*)actuator;
	const TimerCompare* channel = compareFromIOPin(motor->pin);
	const Timer* timer = compareGetTimer(channel);
	uint16_t top = timerGetTOP(timer);

	// New compare threshold
	uint16_t delay=0;

	if( speed > 0 ){
		delay = interpolateU(speed, 0, DRIVE_SPEED_MAX, 0 , top);

		// Just set enable1 high
		pin_make_output(motor->enable1);
		pin_high(motor->enable1);
	}else if(speed < 0){
		delay = interpolateU(speed, 0, DRIVE_SPEED_MIN,  0 , top);
		// Just set enable1 low
		pin_make_output(motor->enable1);
		pin_low(motor->enable1);
	}else{
		// brake
		// Just set enable1 to an input
		pin_make_input(motor->enable1,FALSE);
	}

	// Change the duty cycle
	compareSetThreshold(channel,delay);
}

static void setConnected(__ACTUATOR *actuator, boolean connected){
	MOTOR* motor = (MOTOR*)actuator;
	const TimerCompare* channel = compareFromIOPin(motor->pin);

	// Turn on/off the pin to start/stop sending PWM
	compareSetOutputMode(channel, (connected) ? CHANNEL_MODE_NON_INVERTING : CHANNEL_MODE_DISCONNECT);
}

// Define the class
static __ACTUATOR_DRIVER_CLASS c_motors = MAKE_ACTUATOR_DRIVER_CLASS(&setSpeed,&setConnected);

// Pass the list of servos, the list should be in PROGMEM space to save Flash RAM
// The specified Timer must implement timer compare interrupts and, if so, it will
// ise the timer compare channel A (if there is more than one)
void motorPWMInit(MOTOR_DRIVER* driver){
	// Make sure each servo is set as an output
	for(int i= driver->num_motors-1;i>=0;i--){
		MOTOR* motor = (MOTOR*)pgm_read_word(&driver->motors[i]);

		// Connect motor to driver
		motor->actuator.class = &c_motors;

		// Find the PWM timer compare channel for the output pin
		const TimerCompare* channel = compareFromIOPin(motor->pin);
		if(channel==null){
			setError(PWM_PIN_NOT_AVAILABLE);
			continue;
		}
		if(compareIsInUse(channel)){
			setError(PWM_PIN_IN_USE);
			continue;
		}

		TIMER_MODE mode;
		uint16_t icr;
		uint16_t prescaler;

		const Timer* timer = compareGetTimer(channel);

		// Find the best PWM setting for 10kHz, with 128 steps
		boolean valid = timerCalcPwm(timer, 100000UL, 128, &mode, &icr, &prescaler);

		if(!valid){
			// There is no PWM setting that is valid
			setError( (timerIsInUse(timer)) ? PWM_TIMER_IN_USE : TIMER_HAS_NO_PWM );
		}else{
			// Lets set up the PWM
			if(!timerIsInUse(timer)){
				timerSetMode(timer,mode);
				if(modeIsICR(mode)){
					// Set the ICR
					PORT icrPort = pgm_read_word(&timer->pgm_icr);
					_SFR_MEM16(icrPort)=icr;
				}
			}

			// Make sure the motor pin is an output pin
			pin_make_output(motor->pin);
			pin_low(motor->pin);

			// Make sure the direction pin is set as an output pin
			pin_make_output(motor->enable1);
			pin_low(motor->enable1);

			// Mark the channel as in use
			compareAttach(channel,&nullTimerCompareCallback,0,null);

			// Do this last as it then turns on the timer
			timerSetPrescaler(timer,prescaler);

			// Start off braking
			act_setSpeed(motor,DRIVE_SPEED_BRAKE);

			// Indicate the motor is connected
			act_setConnected(motor,TRUE);
		}
	}
}
