/*
 * $Id$
 *
 * Revision History
 * ================
 * $Log$
 * ===========
 * sdCard.c
 *
 *  Created on: 02-Jul-2009
 *      Author: Clive Webster
 */
#include "sdCard.h"
#include "timer.h"
#include <stdlib.h>

#define DEBUG_SD

#ifdef DEBUG_SD
#include "rprintf.h"
#endif

// Number of retries when expecting an R1
#define WAIT_R1_TIMEOUT 50

/* Transfer-related return codes */
#define TR_OK 			0
#define TR_TIMEOUT     -1
#define TR_ERROR_TOKEN -2
#define TR_NOT_IDLE    -3
#define TR_FAILURE     -4

/* R1 format responses (ORed together as a bit-field) */
#define R1_NOERROR		0
#define R1_IDLE      	1
#define R1_ERASE     	2
#define R1_ILLEGAL   	4
#define R1_CRC_ERR   	8
#define R1_ERASE_SEQ 	16
#define R1_ADDR_ERR  	32
#define R1_PARAM_ERR 	64



/* Common command set */
#define CMD0_GO_IDLE_STATE   			0
#define CMD1_SEND_OPCOND     			1
#define CMD8_SEND_IF_COND        		8
#define CMD9_SEND_CSD        			9
#define CMD10_SEND_CID       			10
#define CMD12_STOP_TRANSMISSION			12
#define CMD13_SEND_STATUS       		13
#define CMD16_SET_BLOCKLEN      		16
#define CMD17_READ_SINGLE_BLOCK 		17
#define CMD18_READ_MULTIPLE_BLOCK		18
#define CMD24_WRITE_BLOCK           	24
#define CMD25_WRITE_MULTIPLE_BLOCK  	25
#define CMD27_PROGRAM_CSD           	27
#define CMD28_SET_WRITE_PROT        	28
#define CMD29_CLR_WRITE_PROT        	29
#define CMD30_SEND_WRITE_PROT       	30
#define CMD32_ERASE_WR_BLK_START_ADDR	32
#define CMD33_ERASE_WR_BLK_END_ADDR   	33
#define CMD38_ERASE                   	38
#define CMD55_APP_CMD                 	55
#define CMD56_GEN_CMD                 	56
#define CMD58_READ_OCR                	58
#define CMD59_CRC_ON_OFF              	59

/* Application-specific commands (always prefixed with CMD55_APP_CMD) */
#define ACMD13_SD_STATUS              	13
#define ACMD22_SEND_NUM_WR_BLOCKS     	22
#define ACMD23_SET_WR_BLK_ERASE_COUNT 	23
#define ACMD41_SEND_OP_COND           	41
#define ACMD42_SET_CLR_CARD_DETECT    	42
#define ACMD51_SEND_SCR               	51

// Format of CSD version 1.0
struct s_csd_v1{
	uint8_t	taac:8;				// data read access time
	uint8_t nsac:8;				// data read access time in CLK cycles (nsac * 100)
	uint8_t tran_speed:8;		// max data transfer rate 32h or 5ah
	uint16_t ccc:12;			// fixed value
	uint8_t read_bl_len:4;		// max read data block length => 2 ^ read_bl_len
	uint8_t read_bl_partial:1;	// is partial blocks for read allowed? = 1 => can write partial blocks
	uint8_t write_blk_misalign:1;// 1 = writes can extend over block boundaries, 0 = cannot
	uint8_t read_blk_misalign:1; // 1 = reads can extend over block boundaries, 0 = cannot
	uint8_t dsr_imp:1; 			// 1=Driver stage implemented on card, 0=not See Chapter 5.5
	uint8_t reserved1:2;		// reserved bits
	uint32_t c_size:12;			// device size
	uint8_t vdd_r_curr_min:3;	// max read current @VDD min
	uint8_t vdd_r_curr_max:3;	// max read current @VDD max
	uint8_t vdd_w_curr_min:3;	// max write current @VDD min
	uint8_t vdd_w_curr_max:3;	// max write current @VDD max
	uint8_t c_size_mult:3;		// device size multiplier. MULT = 2 ^ (c_size_mult + 2)
	uint8_t erase_blk_en:1;		// erase single block enable
	uint8_t sector_size:7;		// erase sector size
	uint8_t wp_grp_size:7;		// write protect group size
	uint8_t wp_grp_enable:1;	// is write protect group enabled?
	uint8_t reserved3:2;		// reserved bits
	uint8_t r2w_factor:3;		// write speed factor
	uint8_t write_bl_len:4;		// max write data block length
	uint8_t write_bl_partial:1;	// is partial blocks for write allowed?
	uint8_t reserved4:5;		// reserved bits
	uint8_t file_format_group:1;// file format group. Read/write.
	uint8_t copy:1;				// copy flag. Read/write.
	uint8_t perm_write_protect:1;// permanent write protect. Read/write.
	uint8_t tmp_write_protect:1; // temporary write protect. Read/write.
	uint8_t file_format:2; 		// file_format
	uint8_t reserved5:2;		// reserved bits
	// followed by 7 bit CRC and single bit=1
};
// Format of CSD version 2.0
struct s_csd_v2{
	uint8_t	taac:8;				// data read access time = 0eh => 1ms
	uint8_t nsac:8;				// data read access time in CLK cycles (nsac * 100)
	uint8_t tran_speed:8;		// max data transfer rate 32h or 5ah
	uint16_t ccc:12;			// fixed value
	uint8_t read_bl_len:4;		// max read data block length = 9 ie 512 bytes
	uint8_t read_bl_partial:1;	// is partial blocks for read allowed? = 0 => No
	uint8_t write_blk_misalign:1;//  = 0
	uint8_t read_blk_misalign:1; //  = 0
	uint8_t dsr_imp:1; 			//
	uint8_t reserved1:6;		// reserved bits
	uint32_t c_size:22;			// device size. Usable memory size = (c_size+1) * 512 Kb
	uint8_t reserved2:1;		// reserved bits
	uint8_t erase_blk_en:1;		// erase single block enable = 1 ie can erase one or more blocks
	uint8_t sector_size:7;		// erase sector size = 7fh. Unused
	uint8_t wp_grp_size:7;		// write protect group size = 0 => Groups not supported
	uint8_t wp_grp_enable:1;	// is write protect group enabled? = 0 => Groups not supported
	uint8_t reserved3:2;		// reserved bits
	uint8_t r2w_factor:3;		// write speed factor = 2. => Write timeout=250ms
	uint8_t write_bl_len:4;		// max write data block length = 9 => 512 bytes
	uint8_t write_bl_partial:1;	// is partial blocks for write allowed? = 0 => not supported
	uint8_t reserved4:5;		// reserved bits
	uint8_t file_format_group:1;// file format group = 0 => Ignored
	uint8_t copy:1;				// copy flag. Read/write.
	uint8_t perm_write_protect:1;// permanent write protect. Read/write.
	uint8_t tmp_write_protect:1; // temporary write protect. Read/write.
	uint8_t file_format:2; 		// file_format = 0. => Ignored
	uint8_t reserved5:2;		// reserved bits
	// followed by 7 bit CRC and single bit=1
};

struct s_csd{
	uint8_t csd_structure:2;	// should contain 01b for version 2, or 00b for version 1
	uint8_t reserved:6;			// reserved bits
	union {
		struct s_csd_v1 v1;
		struct s_csd_v2 v2;
	} csd;
};

struct s_csd_message{
	uint8_t r1;					// The r1 response
	struct s_csd csd;				// The CSD
	uint8_t crc:7;				// The CRC 7 bit checksum
	uint8_t stopbit:1;			// = 1
};

typedef enum e_voltage{
	SD_CMD8_VOLTAGE_2V7_3V6	= 1
} SD_CMD8_VOLTAGE;

static boolean check_r1(uint8_t r1, uint8_t expected){
	if(r1 != expected){
		#ifdef DEBUG_SD
		rprintf("WARNING: R1 status 0x");
		rprintfu08(r1);
		rprintf(", expecting 0x");
		rprintfu08(expected);
		rprintfCRLF();
		#endif
		return FALSE;
	}
	return TRUE;
}



// Pass in the old crc value and the data byte to be added to it
// return the new crc value
//
static uint8_t crc7(uint8_t old_crc, uint8_t data){
	uint8_t new_crc = old_crc;
	for(int x = 7; x >= 0; x--){
		 new_crc <<= 1;
		 if(data & 0x80){
			 new_crc |= 1;
		 }
		 data <<=1;

		 if(new_crc & 0x80){
			 new_crc ^= 0x89; /* XOR with poly -- this clears the top bit, too!*/
		 }
	 }
	 return new_crc;
}

static uint8_t crc7augment(uint8_t old_crc){
	uint8_t new_crc = old_crc;
	for(int x = 0; x < 7; x++){
		new_crc <<= 1;
		if(new_crc & 0x80){
			new_crc ^= 0x89; /* XOR with poly -- this clears the top bit, too!*/
		}
	}

	return new_crc;
}

static uint8_t sdCardXferByte(SD_CARD* card, uint8_t byte){
	return _spiSendByte( __spiDeviceGetBus(card->_device_.bus),byte);
}

// Flush out any pending stuff and attemt to get back to an idle state
// Return TRUE if ok
static boolean sdCardFlushSpi(SD_CARD* card, int timeout){
	uint8_t recv = 0;
    int i = 0;

    /* We unroll the loop to save on variable space. */
    while((i < timeout) && (recv != 0xff))
    {
    	recv = sdCardXferByte(card, 0xff);
        if(recv == 0xff){
        	recv = sdCardXferByte(card, 0xff);
            if(recv == 0xff){
            	recv = sdCardXferByte(card, 0xff);
                if(recv == 0xff){
                	recv = sdCardXferByte(card, 0xff);
                    if(recv == 0xff){
                    	recv = sdCardXferByte(card, 0xff);
                    }
                }
            }
        }
        i++;
    }

    return (recv==0xff) ? TRUE : FALSE;
}

/* Length is # of data bytes to follow R1 response. Status will be in b[0] */
static int sdCardWaitForR1(SD_CARD* card, void * buffer, size_t offset, size_t length){
	uint8_t* b = buffer;
	uint8_t recv = 0xff;
	int i = 0;

	/* Wait for start bit on R1 */
	while(recv & 0x80){
		if(i++ > WAIT_R1_TIMEOUT){
			#ifdef DEBUG_SD
			rprintf("R1 timeout\n");
			#endif
			return TR_TIMEOUT;
		}
		recv = sdCardXferByte(card, 0xff);
	}

	b[offset] = recv; /* Copy in status */

	// Save against the device
	card->response = recv;

	if(length > 0){

		/* Wait for start token on data portion, if any */
		recv = 0xff;
		i = 0;
		while(recv != 0xfe){
			if(i++ > 50){
				#ifdef DEBUG_SD
				rprintf("R1 data timeout\n");
				#endif
				return TR_TIMEOUT;
			}

			recv = sdCardXferByte(card, 0xff);

			if((recv != 0xff) && (recv != 0xfe)){
				/* Not idle bus and not start token, something else. Bus issue?! */
				/* Copy other token into buffer for program to examine */
				b[offset + 1] = recv;
				#ifdef DEBUG_SD
				rprintf("R1 data error 0x");
				rprintfu08(recv);
				rprintfCRLF();
				#endif
				return TR_ERROR_TOKEN;
			}
		}

		/* Read all bytes */
		for(i = 1; i <= length; i++){
			b[offset + i] = sdCardXferByte(card, 0xff);
		}
	}

	/* Eight more bit clocks to finish internal SD operations */
	recv = sdCardXferByte(card, 0xff);
	if(recv != 0xff){
		/* Bus is not idle .. should not happen for single block transfers */
		#ifdef DEBUG_SD
		rprintf("R1 not idle\n");
		#endif
		return TR_NOT_IDLE;
	}

	return TR_OK;
}



// Select the device and send a command
static void sdCardSendCommand(SD_CARD* card, uint8_t command, uint32_t arg){
	uint8_t crc = 0;
	uint8_t byte;

	// Send command
	byte = (command & 0x3f) | 0x40;
	sdCardXferByte(card, byte);
	crc = crc7(crc,byte);

	// Send parameter1
	byte = arg>>24;
	sdCardXferByte(card, byte);
	crc = crc7(crc,byte);

	// Send parameter2
	byte = arg>>16;
	sdCardXferByte(card, byte);
	crc = crc7(crc,byte);

	// Send parameter3
	byte = arg>>8;
	sdCardXferByte(card, byte);
	crc = crc7(crc,byte);

	// Send parameter4
	byte = arg;
	sdCardXferByte(card, byte);
	crc = crc7(crc,byte);

	// Send the checksum which should be 0x95 for GO_IDLE_STATE
	crc = crc7augment(crc);
	byte = (crc<<1) | 1;
	sdCardXferByte(card, byte);

	// wait for response
	// if more than 8 retries, card has timed-out
	// return the received 0xFF
//	int retry=0;
//	while((rtn.r1 = _spiSendByte(bus,0xFF)) == 0xFF){
//		if(retry++ > 8)
//			break;
//	}
//
}

#define RETRIES 10

/* Send the CMD0_GO_IDLE_STATE while CS is asserted */
/* This signals the SD card to fall back to SPI mode */
static boolean sdCardCMD0(SD_CARD* card){
	boolean rtn=FALSE;
	int errors = 0;
	int i;
	uint8_t rxdata;

    while( (rtn==FALSE) && (errors < RETRIES) ){
		#ifdef DEBUG_SD
        rprintf("-> Send CMD0_GO_IDLE_STATE\n");
		#endif
    	// Try to initialis SPI mode
    	sdCardSendCommand(card, CMD0_GO_IDLE_STATE,0);

    	if((i = sdCardWaitForR1(card, &rxdata, 0, 0)) < 0){
            errors++;
            continue;
        }
        if(!check_r1(rxdata,R1_IDLE)){
            errors++;
            continue;
        }

        #ifdef DEBUG_SD
        rprintf("-> Got Idle Card State\n");
		#endif
        rtn = TRUE;
    }
	return rtn;
}

/* Now that card is in SPI mode with IDLE bit set, we need */
/*  to send the command to start its initialization process */
/*  and wait for R1_IDLE to disappear before issuing other commands */
static boolean sdCardACMD41(SD_CARD* card){
    int errors = 0;
    boolean rtn = FALSE;
    while( (rtn==FALSE) && (errors < RETRIES)){
    	uint8_t rxdata;

		#ifdef DEBUG_SD
        rprintf("-> Send CMD55_APP_CMD\n");
		#endif

        sdCardSendCommand(card,CMD55_APP_CMD, 0);
        if(sdCardWaitForR1(card,&rxdata, 0, 0) < 0){
        	/* If this is a MultiMediaCard (not SD), it won't respond here */
			#ifdef DEBUG_SD
        	rprintf("Perhaps this is a MMC card?\n");
			#endif
        	break;
        }
        check_r1(rxdata, R1_IDLE);

		#ifdef DEBUG_SD
		rprintf("-> Send ACMD41_SEND_OP_COND\n");
		#endif

		sdCardSendCommand(card,ACMD41_SEND_OP_COND, 0);
        if(sdCardWaitForR1(card,&rxdata, 0, 0) < 0){
			#ifdef DEBUG_SD
			rprintf("ERROR: on ACMD41_SEND_OP_COND\n");
			#endif
			break;
        }

        if(rxdata & R1_IDLE){
        	delay_cycles(10000);	// let the card start up
        }else{
        	rtn = TRUE;
        }
    }
    return rtn;
}

static boolean sdCardEnableCRC(SD_CARD* card, boolean crc){
	boolean rtn=FALSE;
	uint8_t rxdata;

	/* Enable CRC to protect against bus-induced data errors */
	#ifdef DEBUG_SD
	rprintf("-> Send CMD59_CRC_ON_OFF\n");
	#endif
	sdCardSendCommand(card,CMD59_CRC_ON_OFF, (crc) ? 1 : 0);
	if(sdCardWaitForR1(card,&rxdata, 0, 0) >= 0){
		check_r1(rxdata, R1_NOERROR);
		rtn = TRUE;
	}
	return rtn;
}

/*
static boolean sdCardCMD8(SD_CARD* card, SD_CMD8_VOLTAGE voltage){
	uint8_t rxdata[5];	// 40 bit response code
	boolean rtn = FALSE;
	#ifdef DEBUG_SD
	rprintf("-> Send CMD8_SEND_IF_COND\n");
	#endif
	uint32_t arg = 0xaa;
	arg |= ((uint32_t)voltage)<<8;
	sdCardSendCommand(card,CMD8_SEND_IF_COND, arg);
	if(sdCardWaitForR1(card,rxdata, 0, 4) >=0 ){
		check_r1(rxdata[0],R1_NOERROR);
		if(rxdata[4]!=0xaa){
			#ifdef DEBUG_SD
			rprintf("ERROR: wrong data pattern\n");
			#endif
		}else if((rxdata[3] & 0xf)!=voltage){
			#ifdef DEBUG_SD
			rprintf("ERROR: wrong voltage\n");
			#endif
		}else{
			rtn = TRUE;
		}
	}
	return rtn;
}
*/

static boolean __sdCard__verifyCRC7( void * data, size_t off, size_t len, uint8_t expect ){
	boolean rtn = FALSE;
	uint8_t * b = (uint8_t *)data;

	uint8_t crc = 0;
	int i;

	for(i = 0; i < len; i++)
	{
		crc = crc7(crc, b[off + i]);
	}

	crc = crc7augment(crc);

	if( (crc & 0x7f) == (expect & 0x7f))
	{
		return TRUE;
	}

//	System.out.println("Failed CRC:  Expected: 0x" + toHexString(expect, 4) + " Got: 0x" + toHexString(crc, 4));

	/* Always error out with failure by default. */
	return rtn;
}

// Read the cards CSD data
static boolean sdCardCMD9(SD_CARD* card){
	struct s_csd_message cmd9;

	boolean rtn = FALSE;
	#ifdef DEBUG_SD
	rprintf("-> Send CMD9_SEND_CSD\n");
	#endif

	sdCardSendCommand(card,CMD9_SEND_CSD, 0);
	if(sdCardWaitForR1(card,&cmd9, 0, sizeof(cmd9)-1) >=0 ){
		check_r1(cmd9.r1,R1_NOERROR);

		// Verify the checksum
		if(__sdCard__verifyCRC7( &cmd9, offsetof(struct s_csd_message, csd), sizeof(struct s_csd),cmd9.crc)==FALSE){
			#ifdef DEBUG_SD
			rprintf("CMD9 CRC error");
			#endif
		}else{
			double memsize=0.0;
			card->block_size = 512;
			switch(cmd9.csd.csd_structure){
			case 0:{
				// Version 1.0
				struct s_csd_v1 * v1 = &cmd9.csd.csd.v1;
				uint8_t mult = 1 << (v1->c_size_mult + 2);
				card->block_size = 1 << (v1->read_bl_len);
				double blocknr = (v1->c_size + 1) * mult;
				memsize = blocknr * card->block_size;
				break;
			}
			case 1:{
				// Version 2.0
				struct s_csd_v2 * v2 = &cmd9.csd.csd.v2;
				card->block_size = 1 << (v2->read_bl_len);
				memsize = (v2->c_size + 1) *(512.0 * 1024.0);	// mult of 512kb =>bytes
				break;
			}
			default:
				#ifdef DEBUG_SD
				rprintf("ERROR: CSD version = %d\n",(int)cmd9.csd.csd_structure);
				#endif
				break;
			}
			card->capacityMB = memsize / (1024.0 * 1024.0);		// Size in Mb
			card->num_blocks = memsize / card->block_size; // Max number of blocks

			#ifdef DEBUG_SD
			rprintf("CSD Version %d, Card capacity=",(int)cmd9.csd.csd_structure);rprintfNum(10,16,FALSE,' ',card->capacityMB);rprintf(" Mb blocksize=%d bytes\n",card->block_size);
			#endif
		}
	}else{
		#ifdef DEBUG_SD
		rprintf("ERROR: sending CMD9\n");
		#endif
	}



	return rtn;
}

boolean sdCardReset(SD_CARD* card){

	int rtn = FALSE;

	// Send a few bytes to let it all settle
	for(int i=0;i<10;i++){
		sdCardXferByte(card, 0xff);
	}


	spiDeviceSelect(card, TRUE);	// select the card

    /* Just in case the card is in some unknown state */
    /* Flush out all data and attempt to get back to an idle bus */
    sdCardFlushSpi(card, 516);


    if(sdCardCMD0(card)){
    	if(sdCardACMD41(card)){
    		if(sdCardEnableCRC(card,TRUE)){
    			if(sdCardCMD9(card)){
					#ifdef DEBUG_SD
					rprintf("-> SD_CARD reset\n");
					#endif
					if(card->buffer==null){
						card->buffer = malloc(card->block_size);
					}

					// Card has initialised ok
					if(card->buffer != null){
						rtn = TRUE;
					}else{
						#ifdef DEBUG_SD
						rprintf("-> SD_CARD reset - no RAM avail\n");
						#endif
					}
    			}
    		}
    	}
    }


	spiDeviceSelect(card, FALSE);	// deselect the card

	return rtn;
}


