/*
 * $Id: servos.c,v 1.6 2010/02/06 13:26:04 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: servos.c,v $
 * Revision 1.6  2010/02/06 13:26:04  clivewebster
 * Allow setServoConfig to change the settings for a servo
 *
 * Revision 1.5  2009/10/26 18:53:31  clivewebster
 * *** empty log message ***
 *
 * Revision 1.4  2009/10/09 17:41:29  clivewebster
 * *** empty log message ***
 *
 * ===========
 *
 * servos.c
 *
 * A software PWM driver for between 8 and 10 servos.
 * NB If you have more servos then you can you create multiple servo drivers.
 *
 *
 *  Created on: 15-Mar-2009
 *      Author: Clive Webster
 */
#include "servos.h"
#include "errors.h"
#include "iopin.h"
#include "timer.h"
#include "core.h"


// Called when the next servo needs updating
// Never call it from your own code
static void __servoService(const TimerCompare *timer_compare, void* data){
	uint16_t delay;	// The next delay in cycles
	SERVO_DRIVER* driver = (SERVO_DRIVER*)data;

	if(driver->currentServo < driver->num_servos){
		// turn off the previous channel
		SERVO* servo = (SERVO *)pgm_read_word(&driver->servos[driver->currentServo]);
		pin_low(servo->pin);
	}

	// Start processing the next one
	if(driver->servos_init){
		driver->currentServo++;
	}else{
		driver->servos_init=TRUE;
		// Set the threshold to when the first compare was fired
		// so that if there are multiple banks then each bank is
		// reset to use its actual start time.
		compareSetThreshold(timer_compare,timerGetCounter(compareGetTimer(timer_compare)));
	}

	if(driver->currentServo != driver->num_servos){
		// do the next servo
		if(driver->currentServo > driver->num_servos){
			// restart from the beginning
			driver->currentServo = 0;
			driver->total_delay = 0;
		}

		// Locate the servo
		SERVO* servo = (SERVO *)pgm_read_word(&driver->servos[driver->currentServo]);
		delay = servo->delay;

		// ignore the servo if its not connected
		if(act_isConnected(servo)){
			// turn on servo channel bit
			pin_high(servo->pin);
		}

		// Set time to turn pin low
		compareSetThreshold(timer_compare,delay + compareGetThreshold(timer_compare));

		// Accumulate total time in this pass of the servos
//		driver->total_delay += delay;	// This causes the first servo to get a pulse ever 20ms, but later servos may get them sooner

		// Start servo0 20ms after the start of the last servo
		driver->total_delay = delay;
	}else{
		// pause before going back to the first servo
		delay = (driver->total_delay > driver->servo_cycle) ? 100 :  driver->servo_cycle - driver->total_delay;

		// Re-fire the
		compareSetThreshold(timer_compare,delay + compareGetThreshold(timer_compare));
	}


}

// Callback - for when the speed has been set
static void setSpeed(__ACTUATOR *actuator, DRIVE_SPEED speed){
	SERVO* servo = (SERVO*)actuator;

	// Interpolate the values
	uint16_t ticks=interpolateU(speed, DRIVE_SPEED_MIN, DRIVE_SPEED_MAX, servo->min_ticks , servo->max_ticks);
	if(ticks!=servo->delay){
		CRITICAL_SECTION_START;
		servo->delay = ticks;
		CRITICAL_SECTION_END;
	}
}

// Define the class
static __ACTUATOR_DRIVER_CLASS c_Servos = MAKE_ACTUATOR_DRIVER_CLASS(&setSpeed,null);

// Pass the list of servos, the list should be in PROGMEM space to save Flash RAM
// The specified Timer must implement timer compare interrupts and, if so, it will
// ise the timer compare channel A (if there is more than one)
void servosInit(SERVO_DRIVER* driver, const TimerCompare* channel){

	const Timer* timer = compareGetTimer(channel);

	if(timerIs16bit(timer)){
		int i;

		// Select a prescaler that is bigger than the biggest pause.
		// Added 5ms just in case other interrupts kick in.
		uint16_t prescale;

		// If the timer is already in use then make sure it uses the same mode
		if(timerIsInUse(timer)){
			prescale = timerGetPrescaler(timer);
			if(timerGetMode(timer) != TIMER_MODE_NORMAL){
				setError(PWM_TIMER_IN_USE);
			}
		}else{
			prescale = timerGetBestPrescaler(timer, SERVO_CYCLE + 5 );
		}


		driver->currentServo = driver->num_servos;

		// Find how many timer ticks there will be for SERVO_CYCLE
		uint32_t tmp32 = ticks_per_ms(SERVO_CYCLE,prescale);
		driver->servo_cycle = tmp32;
		// make sure it fits
		if( (tmp32 & 0xFFFF0000UL)!=0){
			setError(SERVO_TIMING);
		}


		// Make sure each servo is set as an output
		for(i= driver->num_servos-1;i>=0;i--){
			SERVO* servo = (SERVO*)pgm_read_word(&driver->servos[i]);

			// Connect the servo to this driver
			servo->actuator.class = &c_Servos;
			servo->driver = driver;
			servo->top = driver->servo_cycle;

			pin_make_output(servo->pin);
			pin_low(servo->pin);

			// Take the useable servo range and calculate the min/max pulses
			// (SERVO_CYCLE*1000) = servo_cycle
			//        min			    x
			// x = (min * g_servo_cycle)/(SERVO_CYCLE*1000)
			/*
			uint32_t min = servo->center_us - servo->range_us;
			uint32_t m1 =  driver->servo_cycle / 1000;
			uint32_t m2 =  m1 * min;
			uint32_t m3 =  m2 / SERVO_CYCLE;
			servo->min_ticks = m3;
			if( (m3 & 0xFFFF0000UL)!=0){
				// doesn't fit
				setError(SERVO_TIMING);
			}

			uint32_t max = servo->center_us + servo->range_us;
			m2 =  m1 * max;
			m3 =  m2 / SERVO_CYCLE;
			servo->max_ticks = m3;
			if( (m3 & 0xFFFF0000UL)!=0){
				// doesn't fit
				setError(SERVO_TIMING);
			}
			*/
			servoSetConfig(servo,servo->center_us,servo->range_us);

			act_setConnected(servo,TRUE);


			// Start off braking
			act_setSpeed(servo,DRIVE_SPEED_BRAKE);
		}


		// Set the threshold to start interrupt chain
		if(getError()==0){
			compareAttach(channel, &__servoService,  compareGetThreshold(channel) + driver->servo_cycle, driver);
		}


		// Do this last as it then turns on the timer
		timerSetPrescaler(timer,prescale);

	}else{
		// We need a 16 bit bit timer
		setError(SERVO_TIMER_MUST_BE_16_BIT);
	}
}

