#ifndef _SERVOS_H_
#define _SERVOS_H_
#include "actuators.h"

/* -------------------------------------------------
 * $Id: servos.h,v 1.4 2010/02/06 13:26:04 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: servos.h,v $
 * Revision 1.4  2010/02/06 13:26:04  clivewebster
 * Allow setServoConfig to change the settings for a servo
 *
 * Revision 1.3  2009/10/09 17:41:29  clivewebster
 * *** empty log message ***
 *
 * ===========
 *
	A servo controller that uses interrupt driven routines
	to control the speed of servos in the background.

	Each driver can cope with up to about 10 servos but you
	can have multiple controllers

#include "servos"

// Define two servos in bank 1
SERVO left  = MAKE_SERVO(FALSE, D1,1500, 500);
SERVO right = MAKE_SERVO(TRUE , D2,1500, 50);
SERVO* PROGMEM servos1[] = {&left,&right};
SERVO_DRIVER bank1 = MAKE_SERVO_DRIVER(servos1);

// Define two servos in bank 2
SERVO top    = MAKE_SERVO(FALSE, D1,1500, 500);
SERVO bottom = MAKE_SERVO(TRUE , D2,1500, 50);
SERVO* PROGMEM servos2[] = {&top,&bottom};
SERVO_DRIVER bank2 = MAKE_SERVO_DRIVER(servos2);

void appInit(void){
	// Initialise the servo controllers
	servosInit(&bank1,TIMER1_COMPAREA);
	servosInit(&bank2,TIMER1_COMPAREB);
}



   ------------------------------------------------- */

/*----------------------------------------------------------------------------------------
*
*	Servo
*
*	Holds all data required by a servo
*
-----------------------------------------------------------------------------------------*/
// Servos need updating every 20ms
#define SERVO_CYCLE 20

// Fwd defs
struct s_servo_driver;

/* -------------------------------------------------------
 * Define the structure for one servo
-------------------------------------------------------- */
typedef struct s_servo{
	__ACTUATOR 			actuator;	// has all the common stuff
	struct s_servo_driver* driver;	// The driver it is attached to, set by init'ing the driver
	const IOPin * 		pin;		// The IO pin it is connected to

	// If you need to change these settings at runtime then use servoSetConfig
	uint16_t			center_us;	// The number of microseconds pulse length to center the servo. On ideal servos it will be 1500. ie 1,5ms
	uint16_t			range_us;	// The number of microseconds either side of center for full range. On ideal servos it be 500. Range = 1ms to 2ms.

	// These are scratch variables used by the driver - DO NOT modify by hand
	uint16_t			min_ticks;	// The no of timer ticks for minimum rotation
	uint16_t			max_ticks;  // The no of timer ticks for maximum rotation
	uint16_t			delay;		// The current delay for the timer
	uint16_t			top;
} SERVO;

// Define the standard constructor for a servo
#define MAKE_SERVO(inverted, iopin, center, range)  { MAKE_ACTUATOR(inverted),null,iopin,center, range, 0,0,0 }

void servoSetConfig(SERVO* servo, uint16_t center, uint16_t range);

/* -------------------------------------------------------
 * Define the data type to store a list of these motors
-------------------------------------------------------- */
typedef SERVO* PROGMEM  SERVO_LIST;

/* -------------------------------------------------------
 * Define the data type to control a list of these motors
-------------------------------------------------------- */
typedef struct s_servo_driver{
	SERVO_LIST* const servos;		// The servos
	uint8_t	currentServo;			// The current servo being worked on
	uint16_t total_delay;			// The total delay so far
	uint8_t num_servos; 			// The total number of servos
	uint16_t servo_cycle;			// The number of timer ticks that represent SERVO_CYCLE
	boolean servos_init;			// Have the servos been initialised
} SERVO_DRIVER;

#define MAKE_SERVO_DRIVER(servolst) { \
	servolst, \
	(uint8_t) 0,\
	(uint16_t)0, \
	(uint8_t)(sizeof(servolst)/sizeof(SERVO*)), \
	(uint16_t)0, \
	FALSE}

/* -------------------------------------------------------
 * Define methods specific to this driver
-------------------------------------------------------- */


// Pass the list of servos, the list should be in PROGMEM space to save Flash RAM
// The specified Timer must implement timer compare interrupts and, if so, it will
// ise the timer compare channel A (if there is more than one)
void servosInit(SERVO_DRIVER* driver, const TimerCompare* channel);
// Initialise the servos using PWM. Each servo needs to be connected to a PWM output pin
void servoPWMInit(SERVO_DRIVER* driver);

// Set all of the servo to be connected/disconnected for this bank of servos
void servosSetConnected(const SERVO_DRIVER* driver, boolean connect);
#define servosConnect(driver) servosSetConnected(driver,TRUE)
#define servosDisconnect(driver) servosSetConnected(driver,FALSE)

// Set all of the servo to the same speed for this bank of servos
void servosSetSpeed(const SERVO_DRIVER* driver, DRIVE_SPEED speed);

#endif


