/*
 * $Id: spi.c,v 1.5 2009/11/02 19:05:46 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: spi.c,v $
 * Revision 1.5  2009/11/02 19:05:46  clivewebster
 * Added revision log
 *
 * ===========
 * spi.c
 *
 *  Created on: 23-Jun-2009
 *      Author: Clive Webster
 *
 *  Implement an SPI interface in hardware
 */
#include "spi.h"
#include <avr/io.h>

// Define the implementations of the virtual classes
static void 	__spiHWInit(_SPI* spi, boolean master);
static void 	__spiHWOff(_SPI* spi);
static void 	__spiHWSetClock(_SPI* spi, SPI_CLOCK clock);
static void 	__spiHWSetDataOrder(_SPI* spi,SPI_DATA_ORDER order);
static void 	__spiHWSetMode(_SPI* spi,SPI_MODE mode);
static uint8_t 	__spiHWSendByte(_SPI* spi, uint8_t data);

SPI_CLASS c_hw_spi = MAKE_SPI_CLASS(&__spiHWInit, &__spiHWOff,&__spiHWSetClock,&__spiHWSetDataOrder, &__spiHWSetMode,&__spiHWSendByte );


#if defined (__AVR_ATmega640__) || defined (__AVR_ATmega2560__)
// Define the SPI pins for ATMega640/ATMega2560
#define SS_PORT PORTB
#define SS_DDR  DDRB
#define SS_PIN  PB0

#define SCK_PORT PORTB
#define SCK_DDR  DDRB
#define SCK_PIN  PB1

#define MOSI_PORT PORTB
#define MOSI_DDR  DDRB
#define MOSI_PIN  PB2

#define MISO_PORT PORTB
#define MISO_DDR  DDRB
#define MISO_PIN  PB3
#elif defined (__AVR_ATmega168__)
// Define the SPI pins for ATMega168
#define SS_PORT PORTB
#define SS_DDR  DDRB
#define SS_PIN  PB2

#define SCK_PORT PORTB
#define SCK_DDR  DDRB
#define SCK_PIN  PB5

#define MOSI_PORT PORTB
#define MOSI_DDR  DDRB
#define MOSI_PIN  PB3

#define MISO_PORT PORTB
#define MISO_DDR  DDRB
#define MISO_PIN  PB4

#elif defined (__AVR_ATmega8__)
// Define the SPI pins for ATMega8
#define SS_PORT PORTB
#define SS_DDR  DDRB
#define SS_PIN  PB2

#define SCK_PORT PORTB
#define SCK_DDR  DDRB
#define SCK_PIN  PB5

#define MOSI_PORT PORTB
#define MOSI_DDR  DDRB
#define MOSI_PIN  PB3

#define MISO_PORT PORTB
#define MISO_DDR  DDRB
#define MISO_PIN  PB4

#elif defined (__AVR_ATmega32__)
// Define the SPI pins for ATMega168
#define SS_PORT PORTB
#define SS_DDR  DDRB
#define SS_PIN  PB4

#define SCK_PORT PORTB
#define SCK_DDR  DDRB
#define SCK_PIN  PB7

#define MOSI_PORT PORTB
#define MOSI_DDR  DDRB
#define MOSI_PIN  PB5

#define MISO_PORT PORTB
#define MISO_DDR  DDRB
#define MISO_PIN  PB6

#elif defined (__AVR_ATmega328P__)
// Define the SPI pins for ATMega328P
#define SS_PORT PORTB
#define SS_DDR  DDRB
#define SS_PIN  PORTB2

#define SCK_PORT PORTB
#define SCK_DDR  DDRB
#define SCK_PIN  PORTB5

#define MOSI_PORT PORTB
#define MOSI_DDR  DDRB
#define MOSI_PIN  PORTB3

#define MISO_PORT PORTB
#define MISO_DDR  DDRB
#define MISO_PIN  PORTB4

#else
#error SPI pins not defined
#endif




//------------- Private methods - dont call directly -----
static void __spiHWInit(_SPI* _spi, boolean master){
	//SPI* spi = (SPI*)_spi;
	CRITICAL_SECTION_START;
    volatile char IOReg;
    if(master){
    	sbi(SCK_DDR, SCK_PIN);		// set SCK as output
    	sbi(MOSI_DDR, MOSI_PIN);	// set MOSI as output
    	sbi(SS_DDR, SS_PIN);		// set SS as output for Master mode to work
        // enable SPI in Master Mode, Data order=MSB first, Mode=0, with SCK = CK/4
        SPCR    = (1<<SPE)|(1<<MSTR);
    }else{
    	sbi(MISO_DDR, MISO_PIN);	// set MISO as output
        SPCR    = (1<<SPE);
    }

    __spiHWSetClock(_spi,_spi->clock);
    __spiHWSetDataOrder(_spi,_spi->order);
    __spiHWSetMode(_spi,_spi->mode);

    // clear SPIF bit in SPSR
    IOReg   = SPSR;
    IOReg   = SPDR;
    CRITICAL_SECTION_END;
}

// Turn off the SPI hardware
static void __spiHWOff(_SPI* spi){
	SPCR &= ~(1<<SPE);
}

static void __spiHWSetDataOrder(_SPI* spi,SPI_DATA_ORDER order){
	if(order == SPI_DATA_ORDER_MSB){
		cbi(SPCR,DORD);
	}else{
		sbi(SPCR,DORD);
	}
}

static void __spiHWSetMode(_SPI* spi,SPI_MODE mode){
	switch(mode){
	case SPI_MODE_0:
		cbi(SPCR,CPOL);
		cbi(SPCR,CPHA);
		break;
	case SPI_MODE_1:
		cbi(SPCR,CPOL);
		sbi(SPCR,CPHA);
		break;
	case SPI_MODE_2:
		sbi(SPCR,CPOL);
		cbi(SPCR,CPHA);
		break;
	case SPI_MODE_3:
		sbi(SPCR,CPOL);
		sbi(SPCR,CPHA);
		break;
	}
}

static void __spiHWSetClock(_SPI* spi, SPI_CLOCK clock){
	cbi(SPCR,SPR1);
	cbi(SPCR,SPR0);
#ifdef SPI2X
	cbi(SPSR,SPI2X);
#endif
	switch(clock){
	case SPI_CLOCK_DIV2:
#ifdef SPI2X
		sbi(SPSR,SPI2X);
#endif
	case SPI_CLOCK_DIV4:
		break;

	case SPI_CLOCK_DIV8:
#ifdef SPI2X
		sbi(SPSR,SPI2X);
#endif
	case SPI_CLOCK_DIV16:
		sbi(SPCR,SPR0);
		break;

	case SPI_CLOCK_DIV32:
#ifdef SPI2X
		sbi(SPSR,SPI2X);
#endif
	case SPI_CLOCK_DIV64:
		sbi(SPCR,SPR1);
		break;

	case SPI_CLOCK_DIV128:
		sbi(SPCR,SPR0);
		sbi(SPCR,SPR1);
		break;
	}
}

static uint8_t __spiHWSendByte(_SPI* spi, uint8_t data){
    SPDR  = data;          			// send Character
    while (!(SPSR & (1<<SPIF)));    // wait until Char is sent
    return SPDR;					// return the received byte
}
