/*
 * $Id: uart.c,v 1.11 2010/01/24 17:39:55 clivewebster Exp $
 *
 * Revision History
 * ================
 * $Log: uart.c,v $
 * Revision 1.11  2010/01/24 17:39:55  clivewebster
 * Moved ISRs out to device,h
 *
 * Revision 1.10  2009/12/11 17:19:29  clivewebster
 * *** empty log message ***
 *
 * Revision 1.9  2009/11/16 03:36:32  clivewebster
 * Fix ATmega8 uart not being included
 *
 * Revision 1.8  2009/11/02 19:07:54  clivewebster
 * Added revision log
 *
 * ===========
 * uart.c
 *
 *  Created on: 17-Mar-2009
 *      Author: Clive Webster
 */
#include "uart.h"

// compatibility for the mega161
#ifndef RXCIE
	#define RXCIE	RXCIE0
	#define TXCIE	TXCIE0
	#define RXEN	RXEN0
	#define TXEN	TXEN0
#endif


// Sets the uart baud rate.
// Argument should be in bits-per-second, like uartSetBaudRate(9600);
// or may be BAUD_RATE_MAX to use the highest possible(if inaccurate) baud rate
void __uartHWSetBaudRate(UART* uart, BAUD_RATE baudrate){
	HW_UART* hwUart = (HW_UART*)uart;
	CRITICAL_SECTION_START;
	// calculate division factor for requested baud rate, and set it
	uint16_t bauddiv = (baudrate==BAUD_RATE_MAX) ? 1 : ((cpu_speed + (baudrate*8L))/(baudrate*16L)-1);
	if(hwUart->u2xMask != 0){
		_SFR_MEM8(hwUart->statusA) &=  ~(hwUart->u2xMask);	// turn off u2x
		uint16_t bauddiv2 = (baudrate==BAUD_RATE_MAX) ? 1 : ((cpu_speed + (baudrate*4L))/(baudrate*8L)-1);
		if(bauddiv2 >= bauddiv){
			_SFR_MEM8(hwUart->statusA) |=  hwUart->u2xMask;	// turn on u2x
			bauddiv = bauddiv2;
		}
	}
	_SFR_MEM8(hwUart->baudL) =  bauddiv;
	_SFR_MEM8(hwUart->baudH) = bauddiv>>8;
	CRITICAL_SECTION_END;
}


// Initializes one UART to the default baud rate
// After running this init function, the processor
// I/O pins that used for uart communications (RXD, TXD)
// are no long available for general purpose I/O.
void __uartHWInit(UART* _uart, BAUD_RATE baud){
	HW_UART* uart = (HW_UART*) _uart;

	// set default baud rate
	uartSetBaudRate(uart, baud);
	// enable RxD/TxD and interrupts
	_SFR_MEM8(uart->statusB) =  BV(RXCIE)|BV(TXCIE)|BV(RXEN)|BV(TXEN);
}

// Disable the uart
static void __uartHWOff(UART * _uart){
	HW_UART* uart = (HW_UART*) _uart;

	_SFR_MEM8(uart->statusB) &=  ~(BV(RXCIE)|BV(TXCIE)|BV(RXEN)|BV(TXEN));

	if(uart->u2xMask != 0){
		_SFR_MEM8(uart->statusA) &=  ~(uart->u2xMask);	// turn off u2x
	}

}



// Sends a single byte over the uart.
void __uartHWStartXmit(UART* uart, uint8_t txData){
	_SFR_MEM8(((HW_UART*)uart)->data) = txData;		// Start sending the character
}



// Create a class with the method overrides for this type of UART
UART_CLASS c_hw_uart = MAKE_UART_CLASS( \
		&__uartHWSetBaudRate, 	\
		&__uartHWStartXmit, 	\
		&__uartHWInit, 			\
		&__uartHWOff,			\
		null,					\
		null					\
		);



